import './conversejs/converse.min.js';
import './conversejs/emojis.js';
import './plugins/converse-rai.js';
import './plugins/sib-mam-history.js'
import { Sib, store, StoreMixin } from 'https://cdn.skypack.dev/@startinblox/core@0.15';
import ComponentPath from './path.js';

class Deferred {
  constructor() {
    this.promise = new Promise(((resolve, reject) => {
      this.resolve = resolve;
      this.reject = reject;
    }).bind(this));
    this.then = this.promise.then.bind(this.promise);
    this.catch = this.promise.catch.bind(this.promise);
  }
}

export const SolidXMPPChat = {
  name: 'solid-xmpp-chat',
  use: [ StoreMixin ],
  attributes: {
    authentication: {
      type: String,
      default: 'login',
    },
    autoLogin: {
      type: Boolean,
      default: true,
    },
    i18n: {
      type: String,
      default: 'fr',
    },
    websocketUrl: {
      type: String,
      default: 'wss://jabber.happy-dev.fr/xmpp-websocket',
    },
  },

  get extra_context() {
    return { "foaf": 'http://xmlns.com/foaf/0.1/', "chatProfile": "http://happy-dev.fr/owl/#chatProfile", "jabberID": "foaf:jabberID" };
  },

  async created() {
    const check_identified = await document.querySelector('sib-auth').getUserIdToken();
    if(check_identified) {
      this.element.attachShadow({
        mode: "open"
      });
      this.importCSS(`${ComponentPath}/dist/conversejs/converse.min.css?min`);
      this.importCSS(`${ComponentPath}/dist/index.css?min`);
      if (window.converse_sib === undefined) {
        this.initializeConverse();
      }
    } // Else, not logged in on page load (even if not on chat)
  },

  importCSS(path) {
    let link = document.createElement('link');
    link.rel = 'stylesheet';
    link.href = path;
    this.element.shadowRoot.append(link);
  },

  empty() {
    this.element.innerHTML = 'Loading...';
  },

  async populate() {
    if(typeof converse_sib !== 'undefined') {
      await converse_sib.loaded_deferred;
      if (this.resource) {
        if (await this.resource.jabberRoom) {
          this.jid = await this.resource['jabberID'];
        } else {
          this.jid = await this.resource['chatProfile.jabberID'];
          const user = await document.querySelector('sib-auth').getUser();
          const userProfile = await store.getData(user['@id'], this.context);
          const contactsURL = await userProfile['contacts.@id'];
          try {
            store.post({
              "contact": this.resource['@id'],
              "@context": this.context
            }, contactsURL);
          } catch (e) {}
        }
        await converse_sib.connected_deferred;
        converse_sib.service.plugins.sibChat.changeChat(
          this.jid,
          await this.resource.jabberRoom,
          this.element.shadowRoot,
        );
        window.dispatchEvent(new CustomEvent('read', {
          detail: {
            resource: {
              "@id": this.resource['@id']
            }
          }
        }));
      }
    } // Else, not logged in, on chat change
  },

  initializeConverse() {
    window.converse_sib = {};

    // Deferred resolved at the end of this function
    converse_sib.loaded_deferred = new Deferred();

    // Deferred resolved after converse.initialize
    converse_sib.connected_deferred = new Deferred();

    // Registering window.converse_sib.service
    converse_sib.service = {
      'waitUntilLoaded': converse_sib.loaded_deferred.promise,
      'initialize': function (options) {
        this.waitUntilLoaded().done(this.api.initialize, options);
      },
      'waitUntilConnected': converse_sib.connected_deferred.promise,
      'plugins': {}
    };

    for(var key in sessionStorage){
        if(sessionStorage.hasOwnProperty(key) && /converse/.test(key)){
            sessionStorage.removeItem(key);
        }
    }
    for(var key in localStorage){
        if(localStorage.hasOwnProperty(key) && /converse/.test(key)){
            localStorage.removeItem(key);
        }
    }
    indexedDB.deleteDatabase('converse-persistent');

    // Alias for solid-xmpp-chat use # want to un-expose converse from window later
    // window.converse_sib.service.api === window.converse
    converse_sib.service.api = converse;

    // Change chat plugin
    converse_sib.service.plugins.sibChat = new(class {
      changeChat(jid, is_groupchat, root) {
        // function isEmptyMessage (attrs) {
        //   if (attrs && attrs.attributes) {
        //       attrs = attrs.attributes;
        //   }
        //   return !attrs['oob_url'] &&
        //       !attrs['file'] &&
        //       !(attrs['is_encrypted'] && attrs['plaintext']) &&
        //       !attrs['message'];
        // };

        // function removeUnnecessaryDayIndicators(view) {
        //   const pred = (el) =>
        //     el.matches('.date-separator') && el.nextElementSibling.matches('.date-separator');
        //   const container = view.el.querySelector('.chat-content__messages');
        //   const to_remove = Array.from(container.children).filter(pred);
        //   to_remove.forEach((el) => el.parentElement.removeChild(el));
        // };
        // function isHidden(classList){
        //   for(let i = 0; i < classList.length; i++){
        //     if(classList[i] === 'hidden'){
        //       return true;
        //     }
        //   }
        //   return false;
        // }
        if (!jid) {
          return;
        }
        jid = jid.toLowerCase();
        this._converse.root = root;

        // Get the current used solid-xmpp-chat
        const converse_el = Array.from(document.querySelectorAll('solid-xmpp-chat'))
          .map(el => el.shadowRoot.getElementById('conversejs'))
          .filter(el => el)
          .pop();

        if (converse_el) {
          root.appendChild(converse_el);
          // if (is_groupchat) {
          //   const jid_to_clear = converse_el.getElementsByClassName('converse-chatboxes');
          //   let room_to_clear_view = '';
          //   if (jid_to_clear.length && jid_to_clear[0].children.length){
          //     for (let i = 0; i < jid_to_clear[0].children.length; i++) {
          //         if(!isHidden(jid_to_clear[0].children[i].classList)){
          //           room_to_clear_view = this._converse.chatboxviews.views[jid_to_clear[0].children[i].id.split('-')[1]];
          //           break;
          //         }
          //     }
          //   }
          //   if (room_to_clear_view) {
          //     if (room_to_clear_view.model.messages.length > 30) {
          //       const non_empty_messages = room_to_clear_view.model.messages.filter((m) => !isEmptyMessage(m));
          //       if (non_empty_messages.length > 30) {
          //       while (non_empty_messages.length > 30) {
          //           non_empty_messages.shift().destroy();
          //       }
          //       removeUnnecessaryDayIndicators(room_to_clear_view);
          //       }
          //     }
          //   }
          // }
        }
        if (is_groupchat) {
          this._converse.api.rooms.open(jid, {}, true);
        } else {
          this._converse.api.chats.open(jid, {}, true);
        }
      }
    });
    // Initialize deferred resolution plugin
    setTimeout(async () => {
      // Initialize change change plugin
      converse.plugins.add('conversejs-changechat', converse_sib.service.plugins.sibChat);

      // Initialize deferred resolution plugin
      converse.plugins.add('conversejs-sib-connected', {
        initialize() {
          this._converse.api.listen.on('connectionInitialized', converse_sib.connected_deferred.resolve);
        }
      });

      // Initialize AUTHFAIL plugin
      converse.plugins.add('conversejs-sib-disconnected', {
        initialize() {
          this._converse.api.listen.on('disconnected', () => {
            if(this._converse.connfeedback.attributes.connection_status === converse.env.Strophe.Status.AUTHFAIL) {
              Array.from(document.querySelectorAll('solid-xmpp-chat')).map(el => el.shadowRoot.innerHTML =
                `<div style='margin:3em;line-height:32px;'><b style='color:red;'>Erreur d'authentification.</b><br /><i style='color:grey;'>Êtes-vous sur le Hubl affilié à votre compte ?</i></div>`
              );
            } else {
              Array.from(document.querySelectorAll('solid-xmpp-chat')).map(el => el.shadowRoot.innerHTML =
                `<div style='margin:3em;line-height:32px;'><b style='color:red;'>Erreur.</b><br /><i style='color:grey;'>${this._converse.connfeedback.attributes.message}</i></div>`
              );
            }
          });
        }
      });

      // Initialize rai plugin
      converse.plugins.add('conversejs-rai', {
        async initialize() {

          let userRooms = (await Promise.all([
            new Promise((resolve, reject) => {
              const circleInterval = setInterval(async () => {
                let retry = false;
                let circles = [];
                const user = await document.querySelector('sib-auth').getUser();
                const userProfile = await store.getData(user['@id'], this.context);
                for (let circleMembership of await userProfile['circles.ldp:contains']) {
                  if (circleMembership) {
                    circles.push(await circleMembership['circle.jabberID']);
                  } else {
                    retry = true;
                  }
                }
                if (!retry) {
                  clearInterval(circleInterval);
                  resolve(circles);
                }
              }, 250);
            }),
            new Promise((resolve, reject) => {
              const projectInterval = setInterval(async () => {
                let retry = false;
                let projects = [];
                const user = await document.querySelector('sib-auth').getUser();
                const userProfile = await store.getData(user['@id'], this.context);
                for (let projectMembership of await userProfile['projects.ldp:contains']) {
                  if (projectMembership) {
                    projects.push(await projectMembership['project.jabberID']);
                  } else {
                    retry = true;
                  }
                }
                if (!retry) {
                  clearInterval(projectInterval);
                  resolve(projects);
                }
              }, 250);
            })
          ])).flat();

          // @MattJ Here userRooms is an array of each jabberID the user is on.
          this._converse.api.trigger('raiRoomsUpdated', userRooms);

          this._converse.api.listen.on('chatRoomActivityIndicators', function (jid) {
            window.dispatchEvent(new CustomEvent('newMessage', {
              detail: {
                jid: jid
              }
            }));
          });
        }
      });

      // Transform hats to custom values
      converse.plugins.add('custom-hats', {
        overrides: {
          getHats: function () {
            const hat_conversions = {'admin': 'Administrateur'};
            const _converse = this;
            const hats = _converse.__super__.getHats.apply(this, arguments);
            if (!hat_conversions) {
              return hats;
            } else {
              const role_affiliations = Object.keys(hat_conversions);
              const custom_hats = hats.map((hat) => {
                if (role_affiliations.includes(hat.title)){
                    return({title: hat_conversions[hat.title]});
                } else {
                    return hat;
                }
              })
              return custom_hats
            }
          }
        }
      });
      //Highly experimental DO NOT USE
//       converse.plugins.add('custom-storage', {
//         overrides: {
//           createStore: function () {
//             const _converse = this;
//             if (arguments.length > 1){
//                 arguments[1] = "none";
//             }
//             return _converse.__super__.createStore.apply(this, arguments);
//           }
//         }
//       });
      //Override converse's request permission with an empty function
      //so that permission request for notifications don't happen
      converse.plugins.add('remove-notifications', {
        overrides: {
          requestPermission: function () {},
          showMessageNotification: function () {},
          showChatStateNotification: function () {},
          showContactRequestNotification: function () {},
          showFeedbackNotification: function () {},
          handleChatStateNotification: function () {},
          handleMessageNotification: function () {},
          handleContactRequestNotification: function () {},
          handleFeedback: function () {},
        }
      });
      // Initialize deferred resolution plugin
      converse.plugins.add('conversejs-sib-focused', {
        initialize() {
          this._converse.api.listen.on('chatBoxFocused', function() {
            // Get the currently used solid-xmpp-chat & send read event of
            const resource = Array.from(document.querySelectorAll('solid-xmpp-chat'))
              .filter(el => {return el.shadowRoot.getElementById('conversejs')})
              .pop()
              .component.resource;
            window.dispatchEvent(new CustomEvent('read', {
              detail: {
                resource: resource
              }
            }));
          });
        }
      });

      const user = await document.querySelector('sib-auth').getUser();
      const userProfile = await store.getData(user['@id'], this.context);
      const jabberID = await userProfile['chatProfile.jabberID'];

      if(this.resource) {
        if (await this.resource.jabberRoom) {
          this.jid = await this.resource['jabberID'];
        } else {
          this.jid = await this.resource['chatProfile.jabberID'];
        }
      } else {
        this.jid = jabberID;
      }

      const idToken = await document.querySelector('sib-auth').getUserIdToken();

      converse.initialize({
        'assets_path': (/skypack/.test(ComponentPath) ? ComponentPath.replace('cdn.skypack.dev', 'unpkg.com') : ComponentPath) + '/dist/conversejs/',
        'authentication': this.element.dataset.authentication || 'login',
        'password': idToken,
        'allow_bookmarks': false,
        'allow_chat_pending_contacts': true,
        'allow_message_retraction': 'moderator',
        'allow_non_roster_messaging': true,
        'allow_dragresize': false,
        'allow_logout': false,
        'archived_messages_page_size': 30,
        'auto_list_rooms': true,
        'auto_login': this.element.dataset.autoLogin === 'true',
        'auto_join_on_invite': true,
        'auto_reconnect': true,
        'auto_register_muc_nickname': false,
        'websocket_url': this.element.dataset.websocketUrl || 'wss://jabber.happy-dev.fr/xmpp-websocket',
        'enable_smacks': true,
        "clear_messages_on_reconnection": false,
        "discover_connection_methods": false,
        'jid': jabberID.toLowerCase(),
        'i18n': this.element.dataset.i18n || 'fr',
        'loglevel': 'fatal',
        'message_archiving': 'always',
        'message_archiving_timeout': 60000,
        'muc_disable_slash_commands': true,
        'muc_hats': ['hats', 'vcard_roles', 'admin'],
        'role_affiliation_hat_conversions': {'admin': 'Administrateur'},
        'muc_nickname_from_jid': false,
        'muc_fetch_members': true,
        'muc_show_info_messages': [],
        'play_sounds': false,
        'root': this.element.shadowRoot,
        'show_client_info': false,
        'show_desktop_notifications': false,
        'persistent_store': 'none',
        'sounds_path': ComponentPath + '/dist/conversejs/',
        'show_send_button': false,
        'view_mode': 'fullscreen',
        'visible_toolbar_buttons': {
          call: false,
          spoiler: false,
          emoji: true,
          fileupload: true, // Not working in current Converse
          toggle_occupants: false
        },
        'whitelisted_plugins': [
          'rai',
          'sib-mam-history',
          'conversejs-sib-disconnected',
          'conversejs-sib-connected',
          'conversejs-sib-focused',
          'conversejs-changechat',
          'conversejs-rai',
          'custom-hats',
          'remove-notifications',
        ],
      });

      converse_sib.loaded_deferred.resolve();
    }, 0);
  }
};

Sib.register(SolidXMPPChat);
